/*
    Copyright 2009-12 Stephan Martin, Dominik Gummel

    This file is part of PID-Boost.

    PID-Boost is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PID-Boost is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PID-Boost.  If not, see <http://www.gnu.org/licenses/>.
*/


/*
Instructions:

1.) Adjust the MAXboost Value below to your personal Limit (3.0 means 3 Bar over Atmospheric pressure)
2.) Upload to your PID Boostcontroller Hardware
3.) While the engine is not running, press the Button on the Board more than 3 Seconds to make a Calibration of the Sensor (LED will turn yellow to indicate Calibration)
4.) Drive around and adjust the Boost with the Potentiometer to your maximum, push the Button short to define this a your maximum allowed boost
4.) Adjust the Boost from 0 to your defined maximum with the Potentiometer.
5.) enjoy!

*/

//do not modify below (only if you realy know what you are doing...)
//-----------------------------------------------------------------------------------------------------------------

//PID Lib:
#include <PID_v1.h>
#include <EEPROM.h>            //for Settings / Max Value storage

//Variables:
unsigned long time;        //the current time
unsigned long pushed;      //Time when the button is pushed
float CalRAWBoost = 0.0;   //the Boost, Absolute!
float CalBoost = 0.0;
float CalLD = 0.0;          //calibration from the Boost
float BatVolt = 0.0;
float MAXboost = 3.0;      //preload with 3.0 Bar
float BoostFactor = (MAXboost+1.0)/4.0;
boolean iButtonState = false;

//PID Lib Stuff, see: http://arduino.cc/playground/Code/PIDLibrary
double Setpoint, SetpointRAW, Input, Output;  //PID Variables
PID myPID(&Input, &Output, &Setpoint,2,5,1, DIRECT); //PID Setup


//Setup:
void setup() {                

  //Set the Pinmodes:
  pinMode(3, OUTPUT);     //(Red Led)
  pinMode(5, OUTPUT);     //(Green Led)
  pinMode(7, INPUT);      //(Onboard Button)  
  digitalWrite(7,HIGH);   //Enable the Pullup on the Onboard Button
  pinMode(9, OUTPUT);     //(N75 Drive)
  
  //Analog Input Overview:
  //Analog In 0 = Boost Input (MPXA6400AP)
  //Analog In 1 = External Potentiometer
  //Analog In 2 = Supply Voltage
  //Analog In 3 = External Switch (unused wire currently, possible AnaIn DigIn DigOut)
  
  //lower the PWM Frequency
  setPwmFrequency(9, 64);
  
  //Read the Boost Calibration back
  CalLD = EEPROMReadDouble(200)/1000.0;      //gets the float back (thats accurate enough)
  
  //Read the maximum Boost Setting:
  MAXboost = EEPROMReadDouble(150)/1000.0;
  //If the Unit is new, this was reading 0, so then modify to 3.0:
  if(MAXboost<=0.01)
  {
    MAXboost = 3.0;
  }
  
  //Read the Supply Voltage
  BatVolt = ((5.0*analogRead(2)/1023.0)*4.09)+0.7;
  
  //Serial On 56k:
  Serial.begin(57600);
  Serial.println("*****************");
  Serial.println("*   PID Boost   *");
  Serial.print("*Supply: ");
  Serial.print(BatVolt);
  Serial.println("V");
  Serial.print("*MaxBoost: ");
  Serial.print(MAXboost);
  Serial.println("Bar");
  Serial.println("*****************");
     
  //Check the Button, if pressed longer than 3sec, reset the MAXboost to 3.0 Bar
  if(digitalRead(7) == LOW)
  {
    //Turn On the RED LED:
    digitalWrite(3, HIGH);
    pushed = millis();
    do
    {
      time = millis();  //Update current time
      delay(100);
      if(time-pushed>=3000)
      {
        //Reset the MAXBoost to 3.0 Bar
        EEPROMWriteDouble(150,3.0*1000);
        MAXboost = 3.0;
        Serial.println("Reset MaxBoost to 3.0 Bar"); 
        digitalWrite(5,HIGH);
        delay(1000);
        digitalWrite(5,LOW);
        //stop this routine
        pushed = time;
      }
    }while (digitalRead(7) == LOW);
  }  

  //Recalc the Boost Setting:
  BoostFactor = (MAXboost+1.0)/4.0;  
     
  //Turn On the Green LED:
  digitalWrite(5, HIGH);
  digitalWrite(3, LOW);
  
  //Init the PID Regulator
  myPID.SetMode(AUTOMATIC);
  myPID.SetSampleTime(200); 
  
}

void loop() {
  
  //Current Time:
  time = millis();
  
  //Fetch the Current Boost level
  Input = analogRead(0);
  
  //And for humans:
  CalRAWBoost = 5.0*Input/1023.0;             //only gets 0-5V
  CalRAWBoost = (CalRAWBoost * 80 - 10)/100;  //makes 0-400Pa out of it
  CalBoost = CalRAWBoost - CalLD;             //apply the offset (ambient pressure)
  
  //Fetch the desired Boost level (smooth & limited)
  SetpointRAW = analogRead(1) * BoostFactor;
  Setpoint = Setpoint * 0.5 + SetpointRAW  * 0.5;
  
  //Compute the PID Value
  myPID.Compute();
  
  //Set the N75 Value
  analogWrite(9,Output);
  
  //Debug Output:
  analogWrite(3,Output);
  
  //Serial Output:
  Serial.print(time);
  Serial.print(";");
  Serial.print(int(Setpoint));
  Serial.print(";");
  Serial.print(int(Input));
  Serial.print(";");
  Serial.print(int(Output));   
  Serial.print(";");
  Serial.println(CalBoost);
  
  
  //small Delay
  delay(10);
  
  //Check the Button, if hold for more than 3sec calibrate LD, if pushed set the current setpoint as MAXboost
  if(digitalRead(7) == LOW)
  {
    //Turn On the RED LED:
    digitalWrite(3, HIGH);
    if(iButtonState == false)
    {
      iButtonState = true;
      pushed = time;
    } else
    {
      //Now, if the diff is more then 3sec, calibrate the LD
      if(time-pushed>=3000)
      {
        CalibrateLD();
        //and reset the State
        iButtonState = false;
       //Turn OFF the RED LED:
       digitalWrite(3, LOW);
       delay(3000);
      }
    }
  } else
  {
    if(iButtonState == true)
    {
      //The Button was only pushed (shorter than 5sec) set the new MAXBoost!
      float NewSetPoint = ((Setpoint / 1025)*4)-1; 
      if(NewSetPoint<=0)
      {
         NewSetPoint = 0.2;
      }
  
      MAXboost = NewSetPoint;
      EEPROMWriteDouble(150,MAXboost*1000);
      BoostFactor = (MAXboost+1.0)/4.0;  
      Serial.print("Set MaxBoost to ");
      Serial.print(NewSetPoint);
      Serial.println(" Bar"); 
      //Turn OFF the RED LED:
      digitalWrite(3, LOW);
      //Turn ON the GREEN LED
      digitalWrite(5, HIGH);
      delay(3000);
      digitalWrite(5, LOW);
    }
    iButtonState = false;
  }
}

//----------------------------------------------------------------------------------------
//Subroutines

void CalibrateLD()
{
      
      //Indicate Calibration:
      //Turn On the RED LED:
      digitalWrite(3, HIGH);
      CalLD = CalRAWBoost;
      //and saved:
      EEPROMWriteDouble(200,CalLD*1000);    //writes the float as long, will do it.   
      Serial.println("Sensor calibrated to ambient pressure."); 
      //Indicate Calibration Finish:
      //Turn OFF the RED LED:
      digitalWrite(3, LOW);
}

//http://www.openobject.org/opensourceurbanism/Storing_Data
void EEPROMWriteDouble(int p_address, long p_value)
{
  byte Byte1 = ((p_value >> 0) & 0xFF);
  byte Byte2 = ((p_value >> 8) & 0xFF);
  byte Byte3 = ((p_value >> 16) & 0xFF);
  byte Byte4 = ((p_value >> 24) & 0xFF);

  EEPROM.write(p_address, Byte1);
  EEPROM.write(p_address + 1, Byte2);
  EEPROM.write(p_address + 2, Byte3);
  EEPROM.write(p_address + 3, Byte4);
}
long EEPROMReadDouble(int p_address)
{
  byte Byte1 = EEPROM.read(p_address);
  byte Byte2 = EEPROM.read(p_address + 1);
  byte Byte3 = EEPROM.read(p_address + 2);
  byte Byte4 = EEPROM.read(p_address + 3);

  long firstTwoBytes = ((Byte1 << 0) & 0xFF) + ((Byte2 << 8) & 0xFF00);
  long secondTwoBytes = (((Byte3 << 0) & 0xFF) + ((Byte4 << 8) & 0xFF00));
  secondTwoBytes *= 65536; // multiply by 2 to power 16 - bit shift 24 to the left

  return (firstTwoBytes + secondTwoBytes);
}
//-------------------------------------------------------------------------------------------------------

void setPwmFrequency(int pin, int divisor) {
  byte mode;
  if(pin == 5 || pin == 6 || pin == 9 || pin == 10) {
    switch(divisor) {
      case 1: mode = 0x01; break;
      case 8: mode = 0x02; break;
      case 64: mode = 0x03; break;
      case 256: mode = 0x04; break;
      case 1024: mode = 0x05; break;
      default: return;
    }
    if(pin == 5 || pin == 6) {
      TCCR0B = TCCR0B & 0b11111000 | mode;
    } else {
      TCCR1B = TCCR1B & 0b11111000 | mode;
    }
  } else if(pin == 3 || pin == 11) {
    switch(divisor) {
      case 1: mode = 0x01; break;
      case 8: mode = 0x02; break;
      case 32: mode = 0x03; break;
      case 64: mode = 0x04; break;
      case 128: mode = 0x05; break;
      case 256: mode = 0x06; break;
      case 1024: mode = 0x7; break;
      default: return;
    }
    TCCR2B = TCCR2B & 0b11111000 | mode;
  }
}

